/*
 * Copyright (c) 2019 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <bluetooth/bluetooth.h>
#include <bluetooth/mesh/models.h>
#include <dk_buttons_and_leds.h>
#include "model_handler.h"

static void led_set(struct bt_mesh_onoff_srv *srv, struct bt_mesh_msg_ctx *ctx,
		    const struct bt_mesh_onoff_set *set,
		    struct bt_mesh_onoff_status *rsp);

static void led_get(struct bt_mesh_onoff_srv *srv, struct bt_mesh_msg_ctx *ctx,
		    struct bt_mesh_onoff_status *rsp);

static const struct bt_mesh_onoff_srv_handlers onoff_handlers = {
	.set = led_set,
	.get = led_get,
};

struct led_ctx {
	struct bt_mesh_onoff_srv srv;
	struct k_work_delayable work;
	uint32_t remaining;
	bool value;
};

static struct led_ctx led_ctx[] = {
#if DT_NODE_EXISTS(DT_ALIAS(led0))
	{ .srv = BT_MESH_ONOFF_SRV_INIT(&onoff_handlers) },
#endif
#if DT_NODE_EXISTS(DT_ALIAS(led1))
	{ .srv = BT_MESH_ONOFF_SRV_INIT(&onoff_handlers) },
#endif
#if DT_NODE_EXISTS(DT_ALIAS(led2))
	{ .srv = BT_MESH_ONOFF_SRV_INIT(&onoff_handlers) },
#endif
#if DT_NODE_EXISTS(DT_ALIAS(led3))
	{ .srv = BT_MESH_ONOFF_SRV_INIT(&onoff_handlers) },
#endif
};

static void led_transition_start(struct led_ctx *led)
{
	int led_idx = led - &led_ctx[0];

	/* As long as the transition is in progress, the onoff
	 * state is "on":
	 */
	dk_set_led(led_idx, true);
	k_work_reschedule(&led->work, K_MSEC(led->remaining));
	led->remaining = 0;
}

static void led_status(struct led_ctx *led, struct bt_mesh_onoff_status *status)
{
	status->remaining_time =
		k_ticks_to_ms_ceil32(
			k_work_delayable_remaining_get(&led->work)) +
		led->remaining;
	status->target_on_off = led->value;
	/* As long as the transition is in progress, the onoff state is "on": */
	status->present_on_off = led->value || status->remaining_time;
}

static void led_set(struct bt_mesh_onoff_srv *srv, struct bt_mesh_msg_ctx *ctx,
		    const struct bt_mesh_onoff_set *set,
		    struct bt_mesh_onoff_status *rsp)
{
	struct led_ctx *led = CONTAINER_OF(srv, struct led_ctx, srv);
	int led_idx = led - &led_ctx[0];

	if (set->on_off == led->value) {
		goto respond;
	}

	led->value = set->on_off;
	if (!bt_mesh_model_transition_time(set->transition)) {
		led->remaining = 0;
		dk_set_led(led_idx, set->on_off);
		goto respond;
	}

	led->remaining = set->transition->time;

	if (set->transition->delay) {
		k_work_reschedule(&led->work, K_MSEC(set->transition->delay));
	} else {
		led_transition_start(led);
	}

respond:
	if (rsp) {
		led_status(led, rsp);
	}
}

static void led_get(struct bt_mesh_onoff_srv *srv, struct bt_mesh_msg_ctx *ctx,
		    struct bt_mesh_onoff_status *rsp)
{
	struct led_ctx *led = CONTAINER_OF(srv, struct led_ctx, srv);

	led_status(led, rsp);
}

static void led_work(struct k_work *work)
{
	struct led_ctx *led = CONTAINER_OF(work, struct led_ctx, work.work);
	int led_idx = led - &led_ctx[0];

	if (led->remaining) {
		led_transition_start(led);
	} else {
		dk_set_led(led_idx, led->value);

		/* Publish the new value at the end of the transition */
		struct bt_mesh_onoff_status status;

		led_status(led, &status);
		bt_mesh_onoff_srv_pub(&led->srv, NULL, &status);
	}
}

/* Set up a repeating delayed work to blink the DK's LEDs when attention is
 * requested.
 */
static struct k_work_delayable attention_blink_work;
static bool attention;

static void attention_blink(struct k_work *work)
{
	static int idx;
	const uint8_t pattern[] = {
#if DT_NODE_EXISTS(DT_ALIAS(led0))
		BIT(0),
#endif
#if DT_NODE_EXISTS(DT_ALIAS(led1))
		BIT(1),
#endif
#if DT_NODE_EXISTS(DT_ALIAS(led2))
		BIT(2),
#endif
#if DT_NODE_EXISTS(DT_ALIAS(led3))
		BIT(3),
#endif
	};

	if (attention) {
		dk_set_leds(pattern[idx++ % ARRAY_SIZE(pattern)]);
		k_work_reschedule(&attention_blink_work, K_MSEC(30));
	} else {
		dk_set_leds(DK_NO_LEDS_MSK);
	}
}

static void attention_on(struct bt_mesh_model *mod)
{
	attention = true;
	k_work_reschedule(&attention_blink_work, K_NO_WAIT);
}

static void attention_off(struct bt_mesh_model *mod)
{
	/* Will stop rescheduling blink timer */
	attention = false;
}

static const struct bt_mesh_health_srv_cb health_srv_cb = {
	.attn_on = attention_on,
	.attn_off = attention_off,
};

static struct bt_mesh_health_srv health_srv = {
	.cb = &health_srv_cb,
};

BT_MESH_HEALTH_PUB_DEFINE(health_pub, 0);

//----------------- @Added ---------------------------------
/* Light Lightness Server */
static void light_set(struct bt_mesh_lightness_srv *srv,
		      struct bt_mesh_msg_ctx *ctx,
		      const struct bt_mesh_lightness_set *set,
		      struct bt_mesh_lightness_status *rsp)
{
}

static void light_get(struct bt_mesh_lightness_srv *srv,
		      struct bt_mesh_msg_ctx *ctx,
		      struct bt_mesh_lightness_status *rsp)
{
}


/* Light CTL */
static void light_ctl_temp_set(struct bt_mesh_light_temp_srv *srv,
			struct bt_mesh_msg_ctx *ctx,
			const struct bt_mesh_light_temp_set *set,
			struct bt_mesh_light_temp_status *rsp)
{
}

static void light_ctl_temp_get(struct bt_mesh_light_temp_srv *srv,
			struct bt_mesh_msg_ctx *ctx,
			struct bt_mesh_light_temp_status *rsp)
{	
}

static void light_ctl_range_update(struct bt_mesh_light_temp_srv *srv,
			struct bt_mesh_msg_ctx *ctx,
			const struct bt_mesh_light_temp_range *old_range,
			const struct bt_mesh_light_temp_range *new_range)
{	
}
static const struct bt_mesh_lightness_srv_handlers ctl_lightness_srv_handlers = {
	.light_set = light_set,
	.light_get = light_get,
};

static const struct bt_mesh_light_temp_srv_handlers ctl_light_temp_handlers = {
	.set = light_ctl_temp_set,
	.get = light_ctl_temp_get,
	.range_update = light_ctl_range_update,
};

static struct bt_mesh_light_ctl_srv light_ctl_srv = 
	BT_MESH_LIGHT_CTL_SRV_INIT(&ctl_lightness_srv_handlers, &ctl_light_temp_handlers);

static struct bt_mesh_lightness_srv lightness_srv =
	BT_MESH_LIGHTNESS_SRV_INIT(&ctl_lightness_srv_handlers);

/* Light LC Server */
static struct bt_mesh_light_ctrl_srv light_ctrl_srv =
	BT_MESH_LIGHT_CTRL_SRV_INIT(&light_ctl_srv.lightness_srv);

/* Scene Server */
static struct bt_mesh_scene_srv scene_srv[7];

/* Generic Property Server */
static struct bt_mesh_prop prop[2] = {0};
static struct bt_mesh_prop_srv prop_srv = BT_MESH_PROP_SRV_INIT((struct bt_mesh_prop *const)&prop, 2, NULL, NULL);
static struct bt_mesh_prop_srv prop_srv1 = BT_MESH_PROP_SRV_INIT((struct bt_mesh_prop *const)&prop, 2, NULL, NULL);
static struct bt_mesh_prop_srv prop_srv2 = BT_MESH_PROP_SRV_INIT((struct bt_mesh_prop *const)&prop, 2, NULL, NULL);
static struct bt_mesh_prop_srv prop_srv3 = BT_MESH_PROP_SRV_INIT((struct bt_mesh_prop *const)&prop, 2, NULL, NULL);
static struct bt_mesh_prop_srv prop_srv4 = BT_MESH_PROP_SRV_INIT((struct bt_mesh_prop *const)&prop, 2, NULL, NULL);



static struct bt_mesh_elem elements[] = {
	BT_MESH_ELEM(
		1, BT_MESH_MODEL_LIST(
			BT_MESH_MODEL_CFG_SRV,
			BT_MESH_MODEL_HEALTH_SRV(&health_srv, &health_pub),
			BT_MESH_MODEL_ONOFF_SRV(&led_ctx[0].srv),
			BT_MESH_MODEL_LIGHT_CTL_SRV(&light_ctl_srv),
			BT_MESH_MODEL_SCENE_SRV(&scene_srv[0]),
			BT_MESH_MODEL_PROP_SRV_USER(&prop_srv),
			BT_MESH_MODEL_PROP_SRV_MFR(&prop_srv),
			BT_MESH_MODEL_PROP_SRV_ADMIN(&prop_srv)
			),
		BT_MESH_MODEL_NONE),

	BT_MESH_ELEM(
		2, BT_MESH_MODEL_LIST(
		BT_MESH_MODEL_ONOFF_SRV(&led_ctx[1].srv),
		BT_MESH_MODEL_LIGHT_TEMP_SRV(&light_ctl_srv.temp_srv),
		BT_MESH_MODEL_PROP_SRV_USER(&prop_srv1),
		BT_MESH_MODEL_PROP_SRV_MFR(&prop_srv1),
		BT_MESH_MODEL_PROP_SRV_ADMIN(&prop_srv1)
		),
		BT_MESH_MODEL_NONE),

	BT_MESH_ELEM(
		3, BT_MESH_MODEL_LIST(
		BT_MESH_MODEL_ONOFF_SRV(&led_ctx[2].srv),
		BT_MESH_MODEL_PROP_SRV_USER(&prop_srv2),
		BT_MESH_MODEL_PROP_SRV_MFR(&prop_srv2),
		BT_MESH_MODEL_PROP_SRV_ADMIN(&prop_srv2)
		),
		BT_MESH_MODEL_NONE),

	BT_MESH_ELEM(
		4, BT_MESH_MODEL_LIST(
		BT_MESH_MODEL_ONOFF_SRV(&led_ctx[3].srv),
		BT_MESH_MODEL_PROP_SRV_USER(&prop_srv3),
		BT_MESH_MODEL_PROP_SRV_MFR(&prop_srv3),
		BT_MESH_MODEL_PROP_SRV_ADMIN(&prop_srv3)
		),
		BT_MESH_MODEL_NONE),

	BT_MESH_ELEM(
		5, BT_MESH_MODEL_LIST(
		BT_MESH_MODEL_ONOFF_SRV(&led_ctx[3].srv),
		BT_MESH_MODEL_LIGHT_CTRL_SRV(&light_ctrl_srv),
		BT_MESH_MODEL_SCENE_SRV(&scene_srv[1])
		),
		BT_MESH_MODEL_NONE),

	BT_MESH_ELEM(
		6, BT_MESH_MODEL_LIST(
		BT_MESH_MODEL_ONOFF_SRV(&led_ctx[3].srv),
		BT_MESH_MODEL_PROP_SRV_USER(&prop_srv4),
		BT_MESH_MODEL_PROP_SRV_MFR(&prop_srv4),
		BT_MESH_MODEL_PROP_SRV_ADMIN(&prop_srv4)
		),
		BT_MESH_MODEL_NONE),



	BT_MESH_ELEM(
		7, BT_MESH_MODEL_LIST(
		BT_MESH_MODEL_ONOFF_SRV(&led_ctx[3].srv),
		BT_MESH_MODEL_HEALTH_SRV(&health_srv, &health_pub),
		BT_MESH_MODEL_LIGHTNESS_SRV(&lightness_srv), 
		BT_MESH_MODEL_SCENE_SRV(&scene_srv[2]),
		BT_MESH_MODEL_PROP_SRV_USER(&prop_srv),
		BT_MESH_MODEL_PROP_SRV_MFR(&prop_srv),
		BT_MESH_MODEL_PROP_SRV_ADMIN(&prop_srv)
		),
		BT_MESH_MODEL_NONE),

	BT_MESH_ELEM(
		8, BT_MESH_MODEL_LIST(
		BT_MESH_MODEL_ONOFF_SRV(&led_ctx[3].srv),
		BT_MESH_MODEL_HEALTH_SRV(&health_srv, &health_pub),
		BT_MESH_MODEL_LIGHTNESS_SRV(&lightness_srv), 
		BT_MESH_MODEL_SCENE_SRV(&scene_srv[3]),
		BT_MESH_MODEL_PROP_SRV_USER(&prop_srv),
		BT_MESH_MODEL_PROP_SRV_MFR(&prop_srv),
		BT_MESH_MODEL_PROP_SRV_ADMIN(&prop_srv)
		),
		BT_MESH_MODEL_NONE),

	BT_MESH_ELEM(
		9, BT_MESH_MODEL_LIST(
		BT_MESH_MODEL_ONOFF_SRV(&led_ctx[3].srv),
		BT_MESH_MODEL_HEALTH_SRV(&health_srv, &health_pub),
		BT_MESH_MODEL_LIGHTNESS_SRV(&lightness_srv), 
		BT_MESH_MODEL_SCENE_SRV(&scene_srv[4]),
		BT_MESH_MODEL_PROP_SRV_USER(&prop_srv),
		BT_MESH_MODEL_PROP_SRV_MFR(&prop_srv),
		BT_MESH_MODEL_PROP_SRV_ADMIN(&prop_srv)
		),
		BT_MESH_MODEL_NONE),

	BT_MESH_ELEM(
		10, BT_MESH_MODEL_LIST(
		BT_MESH_MODEL_ONOFF_SRV(&led_ctx[3].srv),
		BT_MESH_MODEL_HEALTH_SRV(&health_srv, &health_pub),
		BT_MESH_MODEL_LIGHTNESS_SRV(&lightness_srv), 
		BT_MESH_MODEL_SCENE_SRV(&scene_srv[5]),
		BT_MESH_MODEL_PROP_SRV_USER(&prop_srv),
		BT_MESH_MODEL_PROP_SRV_MFR(&prop_srv),
		BT_MESH_MODEL_PROP_SRV_ADMIN(&prop_srv)
		),
		BT_MESH_MODEL_NONE),

	BT_MESH_ELEM(
		11, BT_MESH_MODEL_LIST(
		BT_MESH_MODEL_ONOFF_SRV(&led_ctx[3].srv),
		BT_MESH_MODEL_HEALTH_SRV(&health_srv, &health_pub),
		BT_MESH_MODEL_LIGHTNESS_SRV(&lightness_srv), 
		BT_MESH_MODEL_SCENE_SRV(&scene_srv[6]),
		BT_MESH_MODEL_PROP_SRV_USER(&prop_srv),
		BT_MESH_MODEL_PROP_SRV_MFR(&prop_srv),
		BT_MESH_MODEL_PROP_SRV_ADMIN(&prop_srv)
		),
		BT_MESH_MODEL_NONE),
};

//----------------- End @Added -----------------------------

static const struct bt_mesh_comp comp = {
	.cid = CONFIG_BT_COMPANY_ID,
	.elem = elements,
	.elem_count = ARRAY_SIZE(elements),
};

const struct bt_mesh_comp *model_handler_init(void)
{
	k_work_init_delayable(&attention_blink_work, attention_blink);

	for (int i = 0; i < ARRAY_SIZE(led_ctx); ++i) {
		k_work_init_delayable(&led_ctx[i].work, led_work);
	}

	return &comp;
}
